# module_fm_demodulation.py D. Gibbon, 2021-07-11
# To be used with J_waveform_envelope_F0.py

import numpy as np
from scipy.signal import butter, lfilter, medfilt, tukey

#============================================================
# F0 estimation (adjust if you know what you are doing)

# Signal pre-forming,  F0 post-forming settings

voice = "female"

if voice == "male":
	# Male default settings
	f0min = 100
	f0max = 250
elif voice == "female":
	# Female default settings
	f0min = 100
	f0max = 350
else:
	# General default settings
	f0min = 70
	f0max = 450

f0medfilter = 7
f0tukeyfraction = 0.05

f0framelengthfactor = 0.75			# relative to f0min, > 1
f0frameskipfactor = 0.5			# Default is 1, the frame length

f0diffoffsetlengthfactor = 0.1	# relative fo f0max
f0framedispersion = 0.1			# quasi-noise/voiceless detector - can this work?
f0peakoperation = "median"		# the implmementation of "average"
f0differenceoffset = 0.5

# Hess: track previous F0 values
f0trackupper = 1.3
f0tracklower = 0.7

#==========================
# Centre and peak clipping (adjust if you know what you are doing)

centrethresh = 0.0			# Deals with silence and low volume noise
limitthresh = 0.9

#==========================
# Butterworth bandpass filter (adjust if you know what you are dong)

fmbutterhigh = f0min * 2
fmbutterhighorder = 5
fmbutterlow = f0max
fmbutterloworder = 2

#============================================================

def butterworthfilter(signaldata, cutoff, order, fs, type):
	nyqvist = 0.5 * fs
	normal_cutoff = cutoff / nyqvist
	b, a = butter(order, normal_cutoff, btype=type, analog=False)
	filteredsignal = lfilter(b, a, signaldata)
	return filteredsignal

# butterworthfilter(signal, 150, 5, fs, "lowpass")

#==============================================================

def clipper(sig,thresh,type):
	if type == "centre":
		clipped = (abs(sig) > thresh).astype(int) * sig
	elif type == "limit":
		clipped = (abs(sig) < thresh).astype(int) * sig
	elif type == "lower":
		clipped = (sig > thresh).astype(int) * sig
	elif type == "upper":
		clipped = (sig < thresh).astype(int) * sig
	else:
		print("Unknown type:",type); sys.exit()
	return np.asarray(clipped)

#============================================================
# Applying windowshape to signal window does not improve the result

def f0amdf(signal, fs, windowshape, framestart, framelength, f0diffoffsetlength):

	framestop = framestart + framelength
	framecopy = signal[framestart:framestop]
	framecopydiff = np.diff(framecopy)
	framestd = np.std(framecopydiff)

	if framestd < f0framedispersion:		# anti-noise, quasi-voice-detector
		movingwindowrange = range(framestart+f0diffoffsetlength, framestop)
		meandiffs = [	np.sum(np.abs(framecopy - signal[movwinstart:movwinstart+framelength]))
						for movwinstart in movingwindowrange	]
		meandiffs = list(np.array(meandiffs)/np.max(meandiffs))
		smallestmeandiff = np.min(meandiffs)

		if smallestmeandiff < f0differenceoffset:
			smallestmeandiffpos = meandiffs.index(smallestmeandiff) + f0diffoffsetlength
			period = smallestmeandiffpos / fs
			frequency = 1 / period
		else:
			frequency = 0
	else:
		frequency = 0
	return frequency

def f0movingwindow(signal, fs, windowshape, framelength, frameskip, f0diffoffsetlength):

	f0list = [
		f0amdf(signal, fs, windowshape, framestart, framelength, f0diffoffsetlength)
		for framestart in range(0, len(signal)-3*framelength, frameskip)
		]
	return np.array(f0list)


def f0estimate(signal,fs):

# Frame, offset and window shape definition
	f0frameduration =  1 / f0min
	f0frameduration = f0framelengthfactor * f0frameduration
	framerate = 2 / f0frameduration
	framelength = int(f0frameduration * fs)
	frameskip = int(framelength * f0frameskipfactor)

	f0diffoffsetduration = 1 / f0max	# seconds
	f0diffoffsetlength = int(f0diffoffsetlengthfactor * f0diffoffsetduration * fs)	# samples

	# F0 preprocessing
	signal = clipper(signal,centrethresh,"centre")
	signal = clipper(signal,limitthresh,"limit")
	signal = butterworthfilter(signal, fmbutterlow, fmbutterloworder, fs, "low")
	signal = butterworthfilter(signal, fmbutterhigh, fmbutterhighorder, fs, "high")
	windowshape = tukey(framelength, f0tukeyfraction)

	# F0 estimation loop
	f0array = np.array([
		f0amdf(signal, fs, windowshape, framestart, framelength, f0diffoffsetlength)
		for framestart in range(0, len(signal)-3*framelength, frameskip)
		])
	# F0 median smoothing
	f0array = medfilt(f0array, f0medfilter)

	return f0array, framerate

